# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending
from typing import Optional

from mycroft_bus_client import Message


def request_from_mobile(message: Message) -> bool:
    """
    Check if a request is from a mobile device
    Args:
        message: Message object associated with request

    Returns:
        True if message is from a mobile app, else False
    """
    return message.context.get("mobile", False)


def get_message_user(message: Message) -> Optional[str]:
    """
    Get the user associated with a message
    Args:
        message: Message object associated with request

    Returns:
        Username associated with message
    """
    if not message:
        raise ValueError
    if not hasattr(message, "context"):
        raise AttributeError(type(message))
    return message.context.get("username")
