
configuration_available = {'ttsVoice': {
        "en-au": {"female": "Nicole", "male": "Russell"},
        "en-gb": {"female": "Emma", "male": "Brian"},
        "en-in": {"female": "Raveena", "male": "Aditi"},
        "en-us": {"female": "Joanna", "male": "Joey"},
        "en-gb-wls": {"female": "Geraint", "male": "Geraint"}
        }
}